"""Analysis element for polynomial chaos expansion (PCE).
"""
import logging
import chaospy as cp
import pandas as pd
from easyvvuq import OutputType
from .base import BaseAnalysisElement
from .results import AnalysisResults
from .qmc_analysis import QMCAnalysisResults

__author__ = 'Jalal Lakhlili'
__license__ = "LGPL"

logger = logging.getLogger(__name__)


class PCEAnalysisResults(QMCAnalysisResults):
    implemented = ['sobols_first', 'sobols_total', 'describe']

    def _get_sobols_first(self, qoi, input_):
        """Returns the first order sobol index for a given qoi wrt input variable.

        Parameters
        ----------
        qoi : str
           Quantity of interest
        input_ : str
           Input variable

        Returns
        -------
        float
            First order sobol index.
        """
        raw_dict = AnalysisResults._keys_to_tuples(self.raw_data['sobols_first'])
        return raw_dict[AnalysisResults._to_tuple(qoi)][input_][0]

    def _get_sobols_total(self, qoi, input_):
        """Returns the total order sobol index for a given qoi wrt input variable.

        Parameters
        ----------
        qoi : str
           Quantity of interest
        input_ : str
           Input variable

        Returns
        -------
        float
            Total order sobol index.
        """
        raw_dict = AnalysisResults._keys_to_tuples(self.raw_data['sobols_total'])
        return raw_dict[AnalysisResults._to_tuple(qoi)][input_][0]

    def _get_sobols_first_conf(self, qoi, input_):
        """Not implemented for this method.

        Returns
        -------
        list of floats
            Will return a list with two nans, since this is
        pandas way for handling missing values it seems.
        """
        return [float('nan'), float('nan')]

    def _get_sobols_total_conf(self, qoi, input_):
        """Not implemented for this method.

        Returns
        -------
        list of floats
            Will return a list with two nans, since this is
        pandas way for handling missing values it seems.
        """
        return [float('nan'), float('nan')]

    def describe(self):
        result = {}
        for qoi in self.qois:
            result[qoi] = {
                'count': len(self.samples.axes[0]),
                'mean': self.raw_data['statistical_moments'][qoi]['mean'],
                'std': self.raw_data['statistical_moments'][qoi]['std'],
                'var': self.raw_data['statistical_moments'][qoi]['var'],
                '10%': self.raw_data['percentiles'][qoi]['p10'],
                '90%': self.raw_data['percentiles'][qoi]['p90']
            }
        return pd.DataFrame(result)


class PCEAnalysis(BaseAnalysisElement):

    def __init__(self, sampler=None, qoi_cols=None):
        """Analysis element for polynomial chaos expansion (PCE).

        Parameters
        ----------
        sampler : :obj:`easyvvuq.sampling.pce.PCESampler`
            Sampler used to initiate the PCE analysis
        qoi_cols : list or None
            Column names for quantities of interest (for which analysis is
            performed).
        """

        if sampler is None:
            msg = 'PCE analysis requires a paired sampler to be passed'
            raise RuntimeError(msg)

        if qoi_cols is None:
            raise RuntimeError("Analysis element requires a list of "
                               "quantities of interest (qoi)")

        self.qoi_cols = qoi_cols
        self.output_type = OutputType.SUMMARY
        self.sampler = sampler

    def element_name(self):
        """Name for this element for logging purposes"""
        return "PCE_Analysis"

    def element_version(self):
        """Version of this element for logging purposes"""
        return "0.5"

    def analyse(self, data_frame=None):
        """Perform PCE analysis on input `data_frame`.

        Parameters
        ----------
        data_frame : :obj:`pandas.DataFrame`
            Input data for analysis.

        Returns
        -------
        dict:
            Contains analysis results in sub-dicts with keys -
            ['statistical_moments', 'percentiles', 'sobol_indices',
             'correlation_matrices', 'output_distributions']
        """

        if data_frame is None:
            raise RuntimeError("Analysis element needs a data frame to "
                               "analyse")
        elif data_frame.empty:
            raise RuntimeError(
                "No data in data frame passed to analyse element")

        qoi_cols = self.qoi_cols

        results = {'statistical_moments': {},
                   'percentiles': {},
                   'sobols_first': {k: {} for k in qoi_cols},
                   'sobols_second': {k: {} for k in qoi_cols},
                   'sobols_total': {k: {} for k in qoi_cols},
                   'correlation_matrices': {},
                   'output_distributions': {},
                   }

        # Get the Polynomial
        P = self.sampler.P

        # Get the PCE variante to use (Regression or Projection)
        regression = self.sampler.regression

        # Compute nodes (and weights)
        if regression:
            nodes = cp.generate_samples(order=self.sampler.n_samples,
                                        domain=self.sampler.distribution,
                                        rule=self.sampler.rule)
        else:
            nodes, weights = cp.generate_quadrature(order=self.sampler.polynomial_order,
                                                    dist=self.sampler.distribution,
                                                    rule=self.sampler.rule,
                                                    sparse=self.sampler.quad_sparse,
                                                    growth=self.sampler.quad_growth)

        # Extract output values for each quantity of interest from Dataframe
        samples = {k: [] for k in qoi_cols}
        for run_id in data_frame.run_id.unique():
            for k in qoi_cols:
                data = data_frame.loc[data_frame['run_id'] == run_id][k]
                samples[k].append(data.values)

        # Compute descriptive statistics for each quantity of interest
        for k in qoi_cols:
            # Approximation solver
            if regression:
                if samples[k][0].dtype == object:
                    for i in range(self.sampler.count):
                        samples[k][i] = samples[k][i].astype("float64")
                fit = cp.fit_regression(P, nodes, samples[k], "T")
            else:
                fit = cp.fit_quadrature(P, nodes, weights, samples[k])

            # Statistical moments
            mean = cp.E(fit, self.sampler.distribution)
            var = cp.Var(fit, self.sampler.distribution)
            std = cp.Std(fit, self.sampler.distribution)
            results['statistical_moments'][k] = {'mean': mean,
                                                 'var': var,
                                                 'std': std}

            # Percentiles (Pxx)
            P10 = cp.Perc(fit, 10, self.sampler.distribution)
            P90 = cp.Perc(fit, 90, self.sampler.distribution)
            results['percentiles'][k] = {'p10': P10, 'p90': P90}

            # Sensitivity Analysis: First, Second and Total Sobol indices
            sobols_first_narr = cp.Sens_m(fit, self.sampler.distribution)
            sobols_second_narr = cp.Sens_m2(fit, self.sampler.distribution)
            sobols_total_narr = cp.Sens_t(fit, self.sampler.distribution)
            sobols_first_dict = {}
            sobols_second_dict = {}
            sobols_total_dict = {}
            ipar = 0
            i = 0
            for param_name in self.sampler.vary.get_keys():
                j = self.sampler.params_size[ipar]
                sobols_first_dict[param_name] = sobols_first_narr[i:i + j]
                sobols_second_dict[param_name] = sobols_second_narr[i:i + j]
                sobols_total_dict[param_name] = sobols_total_narr[i:i + j]
                i += j
                ipar += 1
            results['sobols_first'][k] = sobols_first_dict
            results['sobols_second'][k] = sobols_second_dict
            results['sobols_total'][k] = sobols_total_dict

            # Correlation matrix
            results['correlation_matrices'][k] = cp.Corr(
                fit, self.sampler.distribution)

            # Output distributions
            results['output_distributions'][k] = cp.QoI_Dist(
                fit, self.sampler.distribution)
        return PCEAnalysisResults(raw_data=results, samples=data_frame,
                                  qois=self.qoi_cols, inputs=list(self.sampler.vary.get_keys()))



from easyvvuq.analysis.results import AnalysisResults
import pytest


def test_keys_to_tuples():
    assert(AnalysisResults._keys_to_tuples({}) == {})
    assert(AnalysisResults._keys_to_tuples(
        {'a': 1, 'b': 2}) == {('a', 0): 1, ('b', 0): 2})
    assert(AnalysisResults._keys_to_tuples(
        {('a', 0): 1, ('b', 0): 2}) == {('a', 0): 1, ('b', 0): 2})
    assert(AnalysisResults._keys_to_tuples(
        {('a', 0): 1, 'b': 2}) == {('a', 0): 1, ('b', 0): 2})


def test_to_tuple():
    assert(AnalysisResults._to_tuple('a') == ('a', 0))
    assert(AnalysisResults._to_tuple(('a', 0)) == ('a', 0))
    with pytest.raises(RuntimeError):
        AnalysisResults._to_tuple(3)


# def test_not_implemented():
#     results = AnalysisResults()
#     with pytest.raises(AttributeError):
#         results.testing()
#     with pytest.raises(RuntimeError, match=r".* {} .*".format(
#         AnalysisResults.all_analysis_methods)):
#         results.sobols_first()



import os
import easyvvuq as uq
import numpy as np
import chaospy as cp
import pytest
import logging
import pandas as pd
import json
from tests.sc.sobol_model import sobol_g_func
from easyvvuq.analysis.qmc_analysis import QMCAnalysisResults


DESCRIBE_RESULTS = """{"run_id":{"count":400.0,"mean":199.5,"std":115.6143013068,\
"min":0.0,"25%":99.75,"50%":199.5,"75%":299.25,"max":399.0},"x1":{"count":400.0,\
"mean":0.4669087199,"std":0.2900847389,"min":0.0057789696,"25%":0.2085183865,\
"50%":0.4712230161,"75%":0.7243719834,"max":0.9849714587},"x2":{"count":400.0,\
"mean":0.4691844467,"std":0.2925121512,"min":0.0038532032,"25%":0.2018501193,\
"50%":0.4495551186,"75%":0.6869525315,"max":0.9983981908},"f":{"count":400.0,\
"mean":1.0181029471,"std":0.7757795887,"min":0.0170251,"25%":0.3510339027,\
"50%":0.8822882238,"75%":1.5481160285,"max":3.1529538623}}"""


def exact_sobols_g_func(d=2, a=[0.0, 0.5, 3.0, 9.0, 99.0]):
    # for the Sobol g function, the exact (1st-order)
    # Sobol indices are known analytically
    V_i = np.zeros(d)
    for i in range(d):
        V_i[i] = 1.0 / (3.0 * (1 + a[i])**2)
    V = np.prod(1 + V_i) - 1
    logging.debug('Exact 1st-order Sobol indices: ', V_i / V)
    return V_i / V


@pytest.fixture
def data():
    # fix random seed to make this test deterministic
    np.random.seed(10000000)
    # Create the sampler
    vary = {
        "x1": cp.Uniform(0.0, 1.0),
        "x2": cp.Uniform(0.0, 1.0)
    }
    # Select the MC sampler
    mc_sampler = uq.sampling.MCSampler(vary, n_mc_samples=100)
    data = []
    for run_id, sample in enumerate(mc_sampler):
        data.append({
            'run_id': run_id,
            'x1': sample['x1'],
            'x2': sample['x2'],
            'f': sobol_g_func([sample['x1'], sample['x2']], d=2)
        })

    df = pd.DataFrame(data)
    return mc_sampler, df


@pytest.fixture
def results(data):
    # Post-processing analysis
    mc_sampler, df = data
    analysis = uq.analysis.QMCAnalysis(sampler=mc_sampler, qoi_cols=['f'])
    results = analysis.analyse(df)
    return results


def test_results(results):
    assert(isinstance(results, QMCAnalysisResults))
    sobols_first_x1 = results._get_sobols_first('f', 'x1')
    sobols_first_x2 = results._get_sobols_first('f', 'x2')
    sobols_total_x1 = results._get_sobols_total('f', 'x1')
    sobols_total_x2 = results._get_sobols_total('f', 'x2')
    assert(sobols_first_x1 == pytest.approx(0.55690589, 0.001))
    assert(sobols_first_x2 == pytest.approx(0.20727553, 0.001))
    assert(sobols_total_x1 == pytest.approx(0.81327937, 0.001))
    assert(sobols_total_x2 == pytest.approx(0.38049629, 0.001))


def test_results_conf(results):
    sobols_first_x1_conf = results._get_sobols_first_conf('f', 'x1')
    assert(sobols_first_x1_conf[0] == pytest.approx(0.14387, 0.001))
    assert(sobols_first_x1_conf[1] == pytest.approx(0.894288, 0.001))
    sobols_first_x2_conf = results._get_sobols_first_conf('f', 'x2')
    assert(sobols_first_x2_conf[0] == pytest.approx(-0.110633, 0.001))
    assert(sobols_first_x2_conf[1] == pytest.approx(0.467528, 0.001))
    sobols_total_x1_conf = results._get_sobols_total_conf('f', 'x1')
    assert(sobols_total_x1_conf[0] == pytest.approx(0.613689, 0.001))
    assert(sobols_total_x1_conf[1] == pytest.approx(1.018587, 0.001))
    sobols_total_x2_conf = results._get_sobols_total_conf('f', 'x2')
    assert(sobols_total_x2_conf[0] == pytest.approx(0.243612, 0.001))
    assert(sobols_total_x2_conf[1] == pytest.approx(0.492141, 0.001))


def test_full_results(results):
    assert(results.sobols_first().shape == (1, 6))
    assert(results.sobols_total().shape == (1, 6))


def test_describe(results):
    assert(json.loads(results.describe().astype(float).to_json()) == json.loads(DESCRIBE_RESULTS))



import os
import easyvvuq as uq
import numpy as np
import chaospy as cp
import pytest
import logging
import pandas as pd
import math
from tests.sc.sobol_model import sobol_g_func
from easyvvuq.analysis.pce_analysis import PCEAnalysisResults


def exact_sobols_g_func(d=2, a=[0.0, 0.5, 3.0, 9.0, 99.0]):
    # for the Sobol g function, the exact (1st-order)
    # Sobol indices are known analytically
    V_i = np.zeros(d)
    for i in range(d):
        V_i[i] = 1.0 / (3.0 * (1 + a[i])**2)
    V = np.prod(1 + V_i) - 1
    logging.debug('Exact 1st-order Sobol indices: ', V_i / V)
    return V_i / V


@pytest.fixture
def data():
    # fix random seed to make this test deterministic
    np.random.seed(10000000)
    # Create the sampler
    vary = {
        "x1": cp.Uniform(0.0, 1.0),
        "x2": cp.Uniform(0.0, 1.0)
    }
    # Select the MC sampler
    sampler = uq.sampling.PCESampler(vary)
    data = []
    for run_id, sample in enumerate(sampler):
        data.append({
            'run_id': run_id,
            'x1': sample['x1'],
            'x2': sample['x2'],
            'f': sobol_g_func([sample['x1'], sample['x2']], d=2)
        })

    df = pd.DataFrame(data)
    return sampler, df


@pytest.fixture
def results(data):
    # Post-processing analysis
    mc_sampler, df = data
    analysis = uq.analysis.PCEAnalysis(sampler=mc_sampler, qoi_cols=['f'])
    results = analysis.analyse(df)
    return results


def test_results(results):
    assert(isinstance(results, PCEAnalysisResults))
    sobols_first_x1 = results._get_sobols_first('f', 'x1')
    sobols_first_x2 = results._get_sobols_first('f', 'x2')
    sobols_total_x1 = results._get_sobols_total('f', 'x1')
    sobols_total_x2 = results._get_sobols_total('f', 'x2')
    assert(sobols_first_x1 == pytest.approx(0.62644867, 0.001))
    assert(sobols_first_x2 == pytest.approx(0.26789576, 0.001))
    assert(sobols_total_x1 == pytest.approx(0.73210424, 0.001))
    assert(sobols_total_x2 == pytest.approx(0.37355133, 0.001))


def test_results_conf(results):
    sobols_first_x1_conf = results._get_sobols_first_conf('f', 'x1')
    assert(math.isnan(sobols_first_x1_conf[0]))
    assert(math.isnan(sobols_first_x1_conf[1]))
    sobols_first_x2_conf = results._get_sobols_first_conf('f', 'x2')
    assert(math.isnan(sobols_first_x2_conf[0]))
    assert(math.isnan(sobols_first_x2_conf[1]))
    sobols_total_x1_conf = results._get_sobols_total_conf('f', 'x1')
    assert(math.isnan(sobols_total_x1_conf[0]))
    assert(math.isnan(sobols_total_x1_conf[1]))
    sobols_total_x2_conf = results._get_sobols_total_conf('f', 'x2')
    assert(math.isnan(sobols_total_x2_conf[0]))
    assert(math.isnan(sobols_total_x2_conf[1]))


def test_full_results(results):
    assert(results.sobols_first().shape == (1, 6))
    assert(results.sobols_total().shape == (1, 6))


def test_describe(results):
    assert(results.describe().to_dict()['f'] == pytest.approx({
        '10%': 0.009410762945528006,
        '90%': 2.1708835276870935,
        'count': 25.0,
        'mean': 0.91011171024204,
        'std': 0.807805287287411,
        'var': 0.6525493821694965
    }))



"""Represents the results obtained during the analysis stage.
All the analysis classes should implement this in a way that makes
most sense. Provides a more unified interface for accessing the results
in a variety of formats (e.g. NumPy arrays or pandas DataFrames).

Examples
--------

>>> results = PCEAnalysis(sampler, ['a', 'b'])
>>> results.to_pd()
>>> results.describe()
"""

import pandas as pd


class AnalysisResults:
    """Contains the analysis results.

    Parameters
    ----------
    raw_data: obj
        an arbitrary object that contains raw analysis data

    samples: pandas DataFrame
        collated samples
    """

    all_analysis_methods = []
    implemented = []

    def __init__(self, raw_data=None, samples=None, qois=None, inputs=None):
        self.raw_data = raw_data
        self.samples = samples
        self.qois = qois
        self.inputs = inputs

    def __getattr__(self, attr):
        if attr in self.all_analysis_methods:
            raise RuntimeError(
                "analysis results method '{}' is not implement in '{}',\
                 implemented methods are {}".format(
                    attr, self.__class__.__name__, self.implemented))
        else:
            raise AttributeError(
                "type object '{}' has no attribute '{}'".format(self.__class__.__name__, attr))

    def implemented(self):
        """Returns a list of implemented functionality.

        Returns
        -------
        a list of str
           names of types of analysis results that are implemented
           for this method
        """
        return []

    def _get_sobols_first(self, qoi, input_):
        """Returns first order Sobol indices.

        Parameters
        ----------
        qoi - str or tuple
           Quantity of interest or if a tuple quantity of interest plus
        coordinate index (for cases where qoi is vector valued.

        input_ - str
           Input variable name.

        Returns
        -------
        a pandas DataFrame
        """
        raise NotImplementedError

    def _get_sobols_total(self, qoi, input_):
        """Returns total order Sobol indices.

        Returns
        -------
        a pandas DataFrame
        """
        raise NotImplementedError

    def _get_sobols_first_conf(self, qoi, input_):
        """Returns confidence intervals for first order Sobol indices.

        Attributes
        ----------
        qoi : str
            Name of the qoi for which the first order sensitivity index
        confidence interval is requested.
        input_ : str
            Name of the input for which the first order sensitivy index
        confidence interval is requested.

        Returns
        -------
        a pandas DataFrame
        """
        raise NotImplementedError

    def _get_sobols_total_conf(self, qoi, input_):
        """Returns confidence intervals for total order Sobol indices.

        Attributes
        ----------
        qoi : str
            Name of the qoi for which the first order sensitivity index
        confidence interval is requested.
        input_ : str
            Name of the input for which the first order sensitivy index
        confidence interval is requested.

        Returns
        -------
        a pandas DataFrame
        """
        raise NotImplementedError

    def sobols_first(self):
        """Creates a pandas DataFrame with the first order sensitivity
        indices.

        The row indices are of the form (qoi, i) where qoi is the name
        of the quantity of interest and i is the coordinate index
        (will be zero for scalar qois).

        The columns indices are of the form (input, i, {low, est,
        high}) where input is the name of the input variable, i is the
        coordinate provided it is a vector quantitty and the last
        element of the 3-tuple is a string which specifies whether it
        is the estimator value for the first order sensitivity index
        or lower or higher 95% confidence interval values.

        Examples
        --------
        >>> results.sobols_first()
                   x1                           x2
                    0                            0
                  est      high      low       est      high       low
        f 0  0.556906  0.894288  0.14387  0.207276  0.467528 -0.110633

        Returns
        -------
        a pandas DataFrame

        """
        df = {}
        for qoi in self.qois:
            qoi = AnalysisResults._to_tuple(qoi)
            rows = {}
            for input_ in self.inputs:
                key = AnalysisResults._to_tuple(input_)
                key_1 = key + ('low',)
                key_2 = key + ('est',)
                key_3 = key + ('high',)
                rows[key_1] = self._get_sobols_first_conf(qoi, input_)[0]
                rows[key_2] = self._get_sobols_first(qoi, input_)
                rows[key_3] = self._get_sobols_first_conf(qoi, input_)[1]
            df[qoi] = rows
        return pd.DataFrame(df).T

    def sobols_total(self):
        """Creates a pandas DataFrame with the total order sensitivity indices.

        The row indices are of the form (qoi, i) where qoi is the name
        of the quantity of interest and i is the coordinate index
        (will be zero for scalar qois).

        The columns indices are of the form (input, i, {low, est,
        high}) where input is the name of the input variable, i is the
        coordinate provided it is a vector quantitty and the last
        element of the 3-tuple is a string which specifies whether it
        is the estimator value for the first order sensitivity index
        or lower or higher 95% confidence interval values.

        Examples
        --------
        >>> results.sobols_total()
                   x1                            x2
                    0                             0
                  est      high       low       est      high       low
        f 0  0.813279  1.018587  0.613689  0.380496  0.492141  0.243612


        Returns
        -------
        a pandas DataFrame
        """
        df = {}
        for qoi in self.qois:
            qoi = AnalysisResults._to_tuple(qoi)
            rows = {}
            for input_ in self.inputs:
                key = AnalysisResults._to_tuple(input_)
                key_1 = key + ('low',)
                key_2 = key + ('est',)
                key_3 = key + ('high',)
                rows[key_1] = self._get_sobols_total_conf(qoi, input_)[0]
                rows[key_2] = self._get_sobols_total(qoi, input_)
                rows[key_3] = self._get_sobols_total_conf(qoi, input_)[1]
            df[qoi] = rows
        return pd.DataFrame(df).T

    def surrogate(self):
        """Returns the surrogate model as a function from parameter dictionary
        to pandas DataFrame. This only needs to be implemented if the analysis
        method in question provides surrogate models.


        Returns
        -------
        a function that takes a dictionary of parameter - value pairs and returns
        a pandas DataFrame with the results (same output as decoder)
        """
        raise NotImplementedError

    def describe(self, groupby=None, qoi_cols=[], percentiles=[0.1, 0.5, 0.9]):
        """Returns descriptive statistics.

        Examples
        --------
        >>> results.moments()
                   run_id          x1          x2           f
        count  400.000000  400.000000  400.000000  400.000000
        mean   199.500000    0.466909    0.469184    1.018103
        std    115.614301    0.290085    0.292512    0.775780
        min      0.000000    0.005779    0.003853    0.017025
        25%     99.750000    0.208518    0.201850    0.351034
        50%    199.500000    0.471223    0.449555    0.882288
        75%    299.250000    0.724372    0.686953    1.548116
        max    399.000000    0.984971    0.998398    3.152954

        Returns
        -------
        a pandas DataFrame with descriptive statistics
        """
        assert(not self.samples.empty)
        if groupby:
            grouped_data = self.samples.groupby(groupby)
            results = grouped_data.describe(percentiles=percentiles)
            if qoi_cols:
                results = results[qoi_cols]
        else:
            if qoi_cols:
                results = self.samples[qoi_cols].describe()
            else:
                results = self.samples.describe()
        return results

    @staticmethod
    def _keys_to_tuples(dictionary):
        """Convert the keys in the dictionary to tuples.

        Parameters
        ----------
        dictionary : dict
            A dictionary with either strings or tuples as keys.

        Examples
        --------
        >>> AnalysisResults._keys_to_tuples({'a': 1, 'b': 2})
        {('a', 0): 1, ('b', 0): 2})

        >>> AnalysisResults._keys_to_tuples({('a', 0): 1, ('b', 0): 2})
        {('a', 0): 1, ('b', 0): 2})

        >>> AnalysisResults._keys_to_tuples({('a', 0): 1, 'b': 2})
        {('a', 0): 1, ('b', 0): 2})

        Returns
        -------
        A dictionary with tuples as keys.
        """
        new_dict = {}
        for key in dictionary.keys():
            new_dict[AnalysisResults._to_tuple(key)] = dictionary[key]
        return new_dict

    @staticmethod
    def _to_tuple(key):
        """Convert key to tuple if it is string, otherwise leave as is.

        Parameters
        ----------
        key: str or tuple

        Examples
        --------
        >>> AnalysisResults._to_tuple('a')
        ('a', 0)

        >>> AnalysisResults._to_tuple(('a', 0))
        ('a', 0)

        Returns
        -------
        Tuple if key is string, key if key is tuple.
        """
        if isinstance(key, tuple):
            return key
        elif isinstance(key, str):
            return (key, 0)
        else:
            raise RuntimeError("this method expects either a string or tuple")
